/* NU32 Serial Bootloader Application
 * December 2013, Nick Marchuk
 * Windows version
 * make with: gcc nu32utility.c -o nu32utility -lwinmm
 * note: -lwinmm gets access to timeGetTime() from winmm.lib
 * 
 * get a list of available COM ports from Device Manager -> Ports (COM & LPT)
 * use the larger of the two sequential COM ports
 *
 * example: nu32utility \\.\COM34 C:\Users\Nick\MPLABXProjects\Example.x\dist\default\production\Example.x.production.hex 
 * note: on windows, COM ports larger than 9 need the '\\.\' in front of their names
*/

#include <stdio.h>
#include <windows.h> // for serial and time functions
#include <stdlib.h> // for strtol

// special characters
#define SOH 0x01 // start of packet
#define EOT 0x04 // end of packet
#define DLE 0x10 // if SOH or EOT appears in data, preceed it with DLE

// CRC table
int crc_table[16] = {
  0x0000, 0x1021, 0x2042, 0x3063, 0x4084, 0x50a5, 0x60c6, 0x70e7, 
  0x8108, 0x9129, 0xa14a, 0xb16b, 0xc18c, 0xd1ad, 0xe1ce, 0xf1ef
};
int calc_CRC(char * message, int length);

// serial functions
DWORD readFromSerialPort(HANDLE hSerial, char * buffer, int buffersize);
DWORD writeToSerialPort(HANDLE hSerial, char * data, int length);

// call this program with 2 arguments: the name of the serial port, and the location of the hex file
int main(int argc, char *argv[]) {
  printf("\n=== NU32 Serial Bootloader ===\n");
  // check if the right number of args
  if (argc == 3) {
    printf(">Using com port: %s\n", argv[1]);
    printf(">Using hex file at: %s\n", argv[2]);
  }
  else {
    printf("Must be 3 args -> nu32utility COM4 myhex.hex\n");
	printf("=== Bootloader program complete ===\n");
    return(0);
  }
  
  // open serial port
  LPCSTR portname = argv[1];
  DWORD  accessdirection = GENERIC_READ | GENERIC_WRITE;
  HANDLE hSerial = CreateFile(portname,
    accessdirection,
    0,
    0,
    OPEN_EXISTING,
    0,
    0);
  if (hSerial == INVALID_HANDLE_VALUE) {
    printf("Invalid serial port\n");
	printf("=== Bootloader program complete ===\n");
	return(0);
  }
  printf("Opened serial port %s\n",argv[1]);
  
  // set NU32 serial port parameters - 230400 baud, 8-1-n, RTS and CTS
  DCB dcbSerialParams = {0};
  dcbSerialParams.DCBlength=sizeof(dcbSerialParams);
  if (!GetCommState(hSerial, &dcbSerialParams)) {
     printf("Could not get serial port state\n");
  }
  dcbSerialParams.BaudRate=230400;
  dcbSerialParams.ByteSize=8;
  dcbSerialParams.StopBits=ONESTOPBIT;
  dcbSerialParams.Parity=NOPARITY;
  dcbSerialParams.fRtsControl = RTS_CONTROL_HANDSHAKE;
  dcbSerialParams.fOutxCtsFlow = TRUE;
  if(!SetCommState(hSerial, &dcbSerialParams)){
     printf("Some serial port setting error\n");
  }
  // play with these if you get errors (numbers too small)
  // or if performance is too slow (numbers too big)
  COMMTIMEOUTS timeouts={0};
  timeouts.ReadIntervalTimeout=2; // default 50
  timeouts.ReadTotalTimeoutConstant=2; // default 50
  timeouts.ReadTotalTimeoutMultiplier=2; // default 10
  timeouts.WriteTotalTimeoutConstant=2; // default 50
  timeouts.WriteTotalTimeoutMultiplier=2; // default 50
  if(!SetCommTimeouts(hSerial, &timeouts)){
    printf("Some serial timeout setting error\n");
  }  
  
  // function variables
  char array[100];
  int lines = 0;
  int length = 0;
  char * ptr_end;
  unsigned int line;
  int i=0,j=0,k=0;
  char hex[2];
  char message[50]; // SOH+0x03+[HEX]+CRCL+CRCH+EOT 
  char message_buffer[52]; // extra bytes needed for special characters
  int CRC;
  char response_buffer[50];
  char response[50];
  int response_index = 0;
  int response_length = 0;
  int moveOn = 0;
  int percent = 0;
  int startTime;
  float complete;
  int nextPercent = 10;

  // send request for bootloader info, <SOH><0x01><CRCL><CRCH><EOT>
  printf(">Requesting bootloader version information...\n");
  message_buffer[0] = SOH;
  message_buffer[1] = DLE; // hardcoded for the following 0x01 == SOH
  message_buffer[2] = 0x01;
  CRC = calc_CRC(&message_buffer[2],1);
  message_buffer[3] = CRC&0xFF;
  message_buffer[4] = DLE; // hardcoded, the next one is always 16 (DLE)
  message_buffer[5] = (CRC>>8)&0xFF;
  message_buffer[6] = EOT;
  writeToSerialPort(hSerial, message_buffer, 7);
  
  // if bootloader info does not match or timeout, close serial and exit
  // bootloader info response, <SOH><0x01><MajorVer><MinorVer><CRCL><CRCH><EOT>
  moveOn = 0;
  response_index = 0;
  startTime = timeGetTime(); // in ms
  while(!moveOn) {
    response_length = readFromSerialPort(hSerial, response_buffer, 50);
    for(i=0;i<response_length;i++) {
      response[response_index] = response_buffer[i];
	  //printf("got: %d\n",response_buffer[i]);
	  response_index++;
    }
	for(i=0;i<response_index;i++) {
	  if((i>6)&&(response[i]==EOT)&&(response[i-7]==SOH)) { // potential problem here
	    moveOn = 1;
		printf("Bootloader detected, version %d.%d\n",response[i-4],response[i-3]);
		break;
	  }
	}
	// if more than 2s wait, no bootloader is there, leave the program
	if(timeGetTime() - startTime > 2000) {
	  // close the serial port
	  CloseHandle(hSerial);
	  printf("No bootloader detected!\n");
	  printf("Serial port %s closed.\n",argv[1]);
	  // exit
      printf("=== Bootloader program complete ===\n");
      return(0);
	}
  }
  
  // send request to erase flash, <SOH><0x02><CRCL><CRCH><EOT>
  printf(">Requesting flash erase...\n");
  message_buffer[0] = SOH;
  message_buffer[1] = 0x02;
  CRC = calc_CRC(&message_buffer[1],1);
  message_buffer[2] = CRC&0xFF;
  message_buffer[3] = (CRC>>8)&0xFF;
  message_buffer[4] = EOT;
  writeToSerialPort(hSerial, message_buffer, 5);
  
  // wait for response or timeout
  // erase response, <SOH><0x02><CRCL><CRCH><EOT>
  moveOn = 0;
  response_index = 0;
  startTime = timeGetTime(); // in ms
  while(!moveOn) {
    response_length = readFromSerialPort(hSerial, response_buffer, 50);
    for(i=0;i<response_length;i++) {
      response[response_index] = response_buffer[i];
	  response_index++;
    }
	for(i=0;i<response_index;i++) {
	  if((response[i]==EOT)&&(response[i-1]!=DLE)) { // potential problem here
	    moveOn = 1;
		printf("Flash erased!\n");
	  }
	}
	// check to see if 4s has passed (erase takes a long time), quit 
	if(timeGetTime() - startTime > 4000) {
	  // close the serial port
	  CloseHandle(hSerial);
	  printf("Could not erase flash!\n");
	  printf("Serial port %s closed.\n",argv[1]);
	  // exit
      printf("=== Bootloader program complete ===\n");
      return(0);
	}
  }
  
  // try to open file, get number of lines
  FILE *infile;
  infile = fopen(argv[2],"r");
  if(infile == NULL) {
    printf("Could not open file at %s!\n",argv[2]);
    // close the serial port
	CloseHandle(hSerial);
	printf("Serial port %s closed.\n",argv[1]);
	// exit
    printf("=== Bootloader program complete ===\n");
    return(0);
  }
  // loop thru file, get number of hex lines
  while(fscanf(infile,"%s",array) != EOF) {
	lines++;
  }
  printf(">%d lines in file...\n",lines);
  fclose(infile);
  
  // reopen file
  infile = fopen(argv[2],"r");
  if(infile == NULL) {
    printf("Could not open file at %s!\n",argv[2]);
    // close the serial port
	CloseHandle(hSerial);
	printf("Serial port %s closed.\n",argv[1]);
	// exit
    printf("=== Bootloader program complete ===\n");
    return(0);
  }
  
  printf(">Programming...\n");
  printf("||||||||||\n"); // each bar represents 10% complete
  while(percent < lines) {
    // read line and get number of characters in line
	length = 0;
    fscanf(infile,"%s",array);
    while(array[length] != 0) {
      length++;
    }

    // build the message
    j = 0;
    message[j] = SOH;
    j++;
    message[j] = 0x03; // command to write hex line
    j++;
    for (i=1;i<length;i=i+2) { // start at 1 to skip the ':' at beginning of each line
      hex[0] = array[i];
      hex[1] = array[i+1];
      line = strtol(hex, &ptr_end, 16); // unhex the characters
	  message[j] = line;
	  j++;
    }
    CRC = calc_CRC(&message[1],j-1);
    message[j] = CRC&0xFF;
    j++;
    message[j] = (CRC>>8)&0xFF;
    j++;
    message[j] = EOT;
    j++;
  
    // double up for special charaters
    message_buffer[0] = message[0]; // SOH
    message_buffer[1] = message[1]; // 0x03
    k = 2;
    for (i=2;i<j-3;i++) {
      if((message[i]==SOH)||(message[i]==EOT)||(message[i]==DLE)) {
	    message_buffer[k] = DLE;
	    k++;
		message_buffer[k] = message[i];
	  }
	  else {
	    message_buffer[k] = message[i];
	  }
	  k++;
    }
	if ((message[j-3]==SOH)||(message[j-3]==EOT)||(message[j-3]==DLE)) {
	  message_buffer[k] = DLE;
	  k++;
	}	
    message_buffer[k] = message[j-3]; // CRCL
    k++;
	if ((message[j-2]==SOH)||(message[j-2]==EOT)||(message[j-2]==DLE)) {
	  message_buffer[k] = DLE;
	  k++;
	}
    message_buffer[k] = message[j-2]; // CRCH
    k++;
    message_buffer[k] = message[j-1]; // EOT
    k++;
  
    // send message_buffer
    writeToSerialPort(hSerial, message_buffer, k);
  
    // wait for response, <SOH><0x03><CRCL><CRCH><EOT>
    moveOn = 0;
	response_index = 0;
	startTime = timeGetTime(); // in ms
    while(!moveOn) {
	  response_length = readFromSerialPort(hSerial, response_buffer, 50);
      for(i=0;i<response_length;i++) {
        response[response_index] = response_buffer[i];
	    response_index++;
      }
	  for(i=0;i<response_index;i++) {
	    if((response[i]==EOT)&&(response[i-1]!=DLE)) { // potential problem here
	      moveOn = 1;
	    }
	  }
	  // if more than 2s wait, bootloader not responding, leave the program
	  if(timeGetTime() - startTime > 2000) {
	    // close the serial port
	    CloseHandle(hSerial);
	    printf("Could not program!\n");
	    printf("Serial port %s closed.\n",argv[1]);
	    // exit
        printf("=== Bootloader program complete ===\n");
        return(0);
	  }
    }
	percent++;
	complete = (float)percent / (float)lines * 100.0;
	if (complete >= nextPercent) {
	  printf("*"); // show a moving bar of '*' as each line is programmed
	  nextPercent = nextPercent + 10;
	}
  }
  printf("\nProgramming complete!\n");
  
  // close the file
  fclose(infile);
  
  // request jump to application, <SOH><0x05><CRCL><CRCH><EOT>
  printf(">Requesting jump to application...\n");
  message_buffer[0] = SOH;
  message_buffer[1] = 0x05;
  CRC = calc_CRC(&message_buffer[1],1);
  message_buffer[2] = CRC&0xFF;
  message_buffer[3] = (CRC>>8)&0xFF;
  message_buffer[4] = EOT;
  writeToSerialPort(hSerial, message_buffer, 5);
  
  // close serial port
  CloseHandle(hSerial);
  printf("Serial port %s closed.\n",argv[1]);
  
  // exit
  printf("=== Bootloader program complete ===\n");
  return(0);
}

// serial read function
DWORD readFromSerialPort(HANDLE hSerial, char * buffer, int buffersize) {
  DWORD dwBytesRead = 0;
  if(!ReadFile(hSerial, buffer, buffersize, &dwBytesRead, NULL)){
    printf("Some serial read error\n");
  }
  return dwBytesRead;
}

// serial write function
DWORD writeToSerialPort(HANDLE hSerial, char * data, int length) {
  DWORD dwBytesRead = 0;
  if(!WriteFile(hSerial, data, length, &dwBytesRead, NULL)){
    printf("Some serial write error\n");
  }
  return dwBytesRead;
}

// calculate the CRC
int calc_CRC(char * message, int length) {
  int i;
  int crc = 0;
  int j = 0;

  for (j=0;j<length;j++) {
    i = (crc >> 12) ^ (message[j] >> 4);
    crc = crc_table[i & 0x0F] ^ (crc << 4);
    i = (crc >> 12) ^ (message[j] >> 0);
    crc = crc_table[i & 0x0F] ^ (crc << 4);
  } 

  return (crc & 0xFFFF);
}

/* EOF */

